
/*
 * Copyright (C) moto kawasaki <moto@kawasaki3.org>
 */


#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_http.h>


typedef struct {
    ngx_uint_t setfib;        /* fib taken from nginx.conf */
    ngx_uint_t defaultfib;    /* sysctl -n net.my_fibnum */
    ngx_uint_t numfibs;       /* sysctl -n net.fibs */

    ngx_http_upstream_init_pt      original_init_upstream;
    ngx_http_upstream_init_peer_pt original_init_peer;
} ngx_http_upstream_proxy_setfib_srv_conf_t;


static ngx_int_t ngx_http_upstream_init_proxy_setfib_peer(
    ngx_http_request_t *r, ngx_http_upstream_srv_conf_t *us);
static ngx_int_t ngx_http_upstream_init_proxy_setfib(ngx_conf_t *cf,
    ngx_http_upstream_srv_conf_t *us);
static void *ngx_http_upstream_proxy_setfib_create_conf(ngx_conf_t *cf);
static char *ngx_http_upstream_proxy_setfib(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);


static ngx_command_t  ngx_http_upstream_proxy_setfib_commands[] = {

    { ngx_string("proxy_setfib"),
      NGX_HTTP_UPS_CONF|NGX_CONF_TAKE12,
      ngx_http_upstream_proxy_setfib,
      0,
      0,
      NULL },

      ngx_null_command
};


static ngx_http_module_t  ngx_http_upstream_proxy_setfib_module_ctx = {
    NULL,                                  /* preconfiguration */
    NULL,                                  /* postconfiguration */

    NULL,                                  /* create main configuration */
    NULL,                                  /* init main configuration */

    ngx_http_upstream_proxy_setfib_create_conf, /* create server configuration */
    NULL,                                  /* merge server configuration */

    NULL,                                  /* create location configuration */
    NULL                                   /* merge location configuration */
};


ngx_module_t  ngx_http_upstream_proxy_setfib_module = {
    NGX_MODULE_V1,
    &ngx_http_upstream_proxy_setfib_module_ctx, /* module context */
    ngx_http_upstream_proxy_setfib_commands,    /* module directives */
    NGX_HTTP_MODULE,                       /* module type */
    NULL,                                  /* init master */
    NULL,                                  /* init module */
    NULL,                                  /* init process */
    NULL,                                  /* init thread */
    NULL,                                  /* exit thread */
    NULL,                                  /* exit process */
    NULL,                                  /* exit master */
    NGX_MODULE_V1_PADDING
};

static ngx_int_t
ngx_http_upstream_init_proxy_setfib_peer(ngx_http_request_t *r,
    ngx_http_upstream_srv_conf_t *us)
{

    return NGX_OK;
}

static ngx_int_t
ngx_http_upstream_init_proxy_setfib(ngx_conf_t *cf,
    ngx_http_upstream_srv_conf_t *us)
{
    ngx_http_upstream_proxy_setfib_srv_conf_t *pscf;

    pscf = ngx_http_conf_upstream_srv_conf(
               us, ngx_http_upstream_proxy_setfib_module); 

    if (pscf->original_init_upstream(cf, us) != NGX_OK) {
        return NGX_ERROR;
    }

    pscf->original_init_peer = us->peer.init;

    us->peer.init = ngx_http_upstream_init_proxy_setfib_peer;

    return NGX_OK;
}

static void *
ngx_http_upstream_proxy_setfib_create_conf(ngx_conf_t *cf)
{
    ngx_http_upstream_proxy_setfib_srv_conf_t  *conf;
    ngx_uint_t sysctlbuf;
    size_t sysctlsize = sizeof(sysctlbuf);

    conf = ngx_pcalloc(cf->pool,
                       sizeof(ngx_http_upstream_proxy_setfib_srv_conf_t));
    if (conf == NULL) {
        return NULL;
    }

    /*
     * set by ngx_pcalloc():
     *     nothing in terms of setting default value.
     */

    if (sysctlbyname("net.my_fibnum", &sysctlbuf, &sysctlsize, NULL, 0) == 0) {
        conf->defaultfib = sysctlbuf;
    } else {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "reading net.my_fibnum failed");
        return NULL;
    }

    if (sysctlbyname("net.fibs", &sysctlbuf, &sysctlsize, NULL, 0) == 0) {
        conf->numfibs = sysctlbuf;
    } else {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "reading net.fibs failed");
        return NULL;
    }

    conf->setfib = conf->defaultfib;

    return conf;
}


static char *
ngx_http_upstream_proxy_setfib(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_upstream_srv_conf_t               *uscf;
    ngx_http_upstream_proxy_setfib_srv_conf_t  *pscf;

    ngx_int_t    n;
    ngx_str_t   *value;


    uscf = ngx_http_conf_get_module_srv_conf(cf, ngx_http_upstream_module);

    pscf = ngx_http_conf_upstream_srv_conf(uscf,
                                        ngx_http_upstream_proxy_setfib_module);

    if (pscf->original_init_upstream) {
        return "is duplicate";
    }

    pscf->original_init_upstream = uscf->peer.init_upstream
                                   ? uscf->peer.init_upstream
                                   : ngx_http_upstream_init_round_robin;

    uscf->peer.init_upstream = ngx_http_upstream_init_proxy_setfib;

    /* read options */

    value = cf->args->elts;

    n = ngx_atoi(value[1].data, value[1].len);

    if (n == NGX_ERROR || n < 0) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "invalid value \"%V\" in \"%V\" directive",
                           &value[1], &cmd->name);
        return NGX_CONF_ERROR;
    }

ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "NHUPS(): pscf->setfib=%d, pscf->defaultfib=%d, pscf->numfibs=%d", pscf->setfib, pscf->defaultfib, pscf->numfibs);

    pscf->setfib = n;

ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "NHUPS(): pscf->setfib=%d, pscf->defaultfib=%d, pscf->numfibs=%d", pscf->setfib, pscf->defaultfib, pscf->numfibs);

    return NGX_CONF_OK;
}
